/******************************************************************************
*
* CHECK_MYSQL_REPLICATION.C
*
* Program: Mysql plugin for Nagios
* License: GPL
* Copyright (c) 2002 Dave Viner (dviner@apache.org)
*  
* 
* $Id: check_mysql.c,v 1.2 2002/11/12 11:26:01 kdebisschop Exp $
*
* Description:
*
* This plugin is for testing the replication between a master and a slave
* MySQL database.
******************************************************************************/

#define PROGNAME "check_mysql_replication"
#define REVISION "$Revision: 1.2 $"
#define COPYRIGHT "1999-2002"

#include "common.h"
#include "utils.h"

#include <mysql/mysql.h>
#include <mysql/errmsg.h>

char *db_user = "";
char *masterdb_host = "";
char *slavedb_host = "";
char *db_pass = "";
char *db = "";
int limit = 10000;
unsigned int db_port = MYSQL_PORT;
int verbose = FALSE;

char *slave_sql =  "SHOW SLAVE STATUS";
char *master_sql = "SHOW MASTER STATUS";

int process_arguments (int, char **);
int validate_arguments (void);
int check_disk (int usp, int free_disk);
void print_help (void);
void print_usage (void);

int
main (int argc, char **argv)
{
	
	MYSQL master_mysql, slave_mysql;
	MYSQL_RES *master_state, *slave_state;
	MYSQL_FIELD *master_field, *slave_field;
	MYSQL_ROW row;
	int num_fields, i, slave_version;
	int master_position, slave_position;
	int slave_running = 0;
	int slave_io_running = 0;
	int slave_sql_running = 0;
	int slave_last_errno = 0;
	char slave_last_error[MAX_INPUT_BUFFER] = "";

	int result = STATE_UNKNOWN;
	
	if (process_arguments (argc, argv) != OK)
		usage ("Invalid command arguments supplied\n");
	

	if (verbose)
	{
		printf("Master: %s\nSlave: %s\nUser: %s\n"
			"Pass: %s\nLimit %i\n", 
			masterdb_host, slavedb_host, db_user, db_pass, limit);
	}
	


	/* initialize mysql  */
	mysql_init (&master_mysql);
	mysql_init (&slave_mysql);
	
	/* establish a connection to the master server */
	if (!mysql_real_connect
		(&master_mysql, masterdb_host, db_user, db_pass, NULL, db_port, NULL, 0)) {
		printf ("%s\n", mysql_error(&master_mysql));
		return STATE_CRITICAL;		
	}
	
	/* establish a connection to the slave server */
	if (!mysql_real_connect
		(&slave_mysql, slavedb_host, db_user, db_pass, NULL, db_port, NULL, 0)) {
		printf ("%s\n", mysql_error(&slave_mysql));
		return STATE_CRITICAL;		
	}

	/* query the master server and fetch the result */
	if (mysql_query (&master_mysql, master_sql))
	{
		printf("Error querying master db %s\n", mysql_error(&master_mysql));
		return STATE_CRITICAL;
	}
	master_state = mysql_store_result (&master_mysql);

	/* query the slave server and fetch the result */
	if (mysql_query (&slave_mysql, slave_sql))
	{
		printf("Error querying master db %s\n", mysql_error(&master_mysql));
		return STATE_WARNING;
	}
	slave_state = mysql_store_result (&slave_mysql);

	/* make sure that we got result values */
	if (!master_state  || !slave_state)
	{
		printf("Nothing returned from either the master query or the slave query\n");
		return STATE_UNKNOWN;
	}

	/* find the status of the master server */
	num_fields = mysql_num_fields (master_state);
	master_field = mysql_fetch_fields (master_state);
	while ((row = mysql_fetch_row (master_state)))
	{
		long *lengths;
		lengths = mysql_fetch_lengths (master_state);
		for (i=0; i < num_fields; i++)
		{
			if (verbose)
				printf("MASTER: %s => [%.*s]\n", 
					master_field[i].name, (int) lengths[i], 
					row[i] ? row[i] : "NULL");

			/* save aside the position of the master */
			if (strcmp(master_field[i].name,"Position") == 0)
				master_position = row[i] ? atoi (row[i]) : 0;
		}
		if (verbose)
			printf("\n\n");
	}

	/* find the status of the slave server */
	slave_version = atoi ( mysql_get_server_info (&slave_mysql) );
	if (verbose)
		printf("Slave version %i\n", slave_version);
	num_fields = mysql_num_fields (slave_state);
	slave_field = mysql_fetch_fields (slave_state);
	while ((row = mysql_fetch_row (slave_state)))
	{
		long *lengths;
		lengths = mysql_fetch_lengths (slave_state);
		for (i=0; i < num_fields; i++)
		{
			if (verbose)
				printf("SLAVE: %s => [%.*s]\n", 
					slave_field[i].name, (int) lengths[i], 
					row[i] ? row[i] : "NULL");

			/* first look for the last errno and last error from the slave */
			if (strcmp(slave_field[i].name,"Last_errno") == 0)
				slave_last_errno = atoi (row[i]);
			if (strcmp(slave_field[i].name,"Last_error") == 0)
				strncpy (slave_last_error, row[i], MAX_INPUT_BUFFER);

			/* now look for the slave position and whether its running.
			* this changed between mysql version 3.x and 4.x.  Version 3
			* has Slave_Running, and Pos.  But version 4 has Slave_IO_running,
			* Slave_SQL_Running and Read_Master_Log_Pos.  I'm not sure why 
			* it was changed, but this chunk of code checks for that.
			*/
			if (slave_version > 3 )
			{
				if (strcmp(slave_field[i].name,"Read_Master_Log_Pos") == 0)
					slave_position = row[i] ? atoi (row[i]) : 0;
				else if (strcmp(slave_field[i].name,"Slave_IO_Running") == 0)
				{
					if (row[i] && (strcmp(row[i],"Yes") == 0))
						slave_io_running = 1;
					else
						slave_io_running = 0;
				}
				else if (strcmp(slave_field[i].name,"Slave_SQL_Running") == 0)
				{
					if (row[i] && (strcmp(row[i],"Yes") == 0))
						slave_sql_running = 1;
					else 
						slave_sql_running = 0;
				}
			}
			else
			{
				if (strcmp(slave_field[i].name,"Pos") == 0)
					slave_position = row[i] ? atoi (row[i]) : 0;
				else if (strcmp(slave_field[i].name,"Slave_Running") == 0)
					if (row[i] && (strcmp(row[i],"Yes") == 0))
						slave_running = 1;
					else
						slave_running = 0;
			}
		}
		if (verbose)
			printf("\n\n");
	}

	if (verbose)
		printf ("Found master position %i Slave position %i threshold %i\n", 
			master_position, slave_position, limit);

	/* 
	* Now make version 4's slave running stuff look like version 3
	* This makes it easier to handle later.
	*/
	if (slave_version > 3)
	{
		if ((slave_io_running == 1) && (slave_sql_running == 1))
			slave_running = 1;
		else
			slave_running = 0;
	}

	/* so now, slave_running tells us if the slave is running, and 
	* the positions are recorded in master_pos and slave_pos
	*/
	if (slave_running == 0)
	{
		printf("Slave not running\n");
		result = STATE_CRITICAL;
		
	}
	else
	{
		/* look at the last errno from the slave query */
		if (slave_last_errno != 0)
		{
			printf("Slave error errno %i error %s\n", 
				slave_last_errno, slave_last_error);
			result = STATE_WARNING;
		}
		else
		{
			/* compare the slave position and the master position against
			* the limit requested 
			*/
			if ((master_position - slave_position) > limit)
			{
				printf("Replication slow to slave %s\n", slavedb_host);
				result = STATE_WARNING;
			}
			else
			{
				if (verbose)
					printf("Replication ok\n");
				result = STATE_OK;
			}
		}
	}

	/* free the result sets */
	mysql_free_result (master_state);
	mysql_free_result (slave_state);
	/* close the connection */
	mysql_close (&master_mysql);
	mysql_close (&slave_mysql);
	
	
	return result;
}





/* process command-line arguments */
int
process_arguments (int argc, char **argv)
{
	int c;
	
#ifdef HAVE_GETOPT_H
	int option_index = 0;
	static struct option long_options[] = {
		{"master_hostname", required_argument, 0, 'H'},
		{"slave_hostname", required_argument, 0, 'S'},
		{"username", required_argument, 0, 'u'},
		{"password", required_argument, 0, 'p'},
		{"port", required_argument, 0, 'P'},
		{"limit", required_argument, 0, 'l'},
		{"verbose", no_argument, 0, 'v'},
		{"version", no_argument, 0, 'V'},
		{"help", no_argument, 0, 'h'},
		{0, 0, 0, 0}
	};
#endif
	
	if (argc < 1)
		return ERROR;
	
	while (1) {
#ifdef HAVE_GETOPT_H
		c =
			getopt_long (argc, argv, "hVvP:p:u:d:H:S:l:", long_options, &option_index);
#else
		c = getopt (argc, argv, "hVvP:p:u:d:H:S:l:");
#endif
		
		if (c == -1 || c == EOF)
			break;
		
		switch (c) {
		case 'H':									/* hostname */
			if (is_host (optarg)) {
				masterdb_host = optarg;
			}
			else {
				usage ("Invalid host name\n");
			}
			break;
		case 'd':									/* hostname */
			db = optarg;
			break;
		case 'u':									/* username */
			db_user = optarg;
			break;
		case 'p':									/* authentication information: password */
			db_pass = optarg;
			break;
		case 'P':									/* critical time threshold */
			db_port = atoi (optarg);
			break;
		case 'S':
			slavedb_host = optarg;
			break;
		case 'l':
			limit = atoi (optarg);
			break;
		case 'v':
			verbose = TRUE;
			break;
		case 'V':									/* version */
			print_revision (PROGNAME, REVISION);
			exit (STATE_OK);
		case 'h':									/* help */
			print_help ();
			exit (STATE_OK);
		case '?':									/* help */
			usage ("Invalid argument\n");
		}
	}
	
	c = optind;
	
	while ( argc > c ) {
		if (strlen(masterdb_host) == 0)
		{
			if (is_host (argv[c])) {
				masterdb_host = argv[c++];
			}
			else {
				usage ("Invalid master host name");
			}
		}
		else if (strlen(slavedb_host) == 0)
		{
			if (is_host(argv[c]))
				slavedb_host = argv[c++];
			else
				usage ("Invalid slave host name");
		}
		else if (strlen(db_user) == 0)
			db_user = argv[c++];
		else if (strlen(db_pass) == 0)
			db_pass = argv[c++];
		else if (strlen(db) == 0)
			db = argv[c++];
		else if (is_intnonneg (argv[c]))
			db_port = atoi (argv[c++]);
		else 
			break;
	}
	return validate_arguments ();
}





int
validate_arguments (void)
{
	return OK;
}





void
print_help (void)
{
	print_revision (PROGNAME, REVISION);
	printf
		("Copyright (c) 2000 Didi Rieder/Karl DeBisschop\n\n"
		"This plugin is for testing a mysql server.\n");
	print_usage ();
	printf
		("\nThere are no required arguments. By default, the local database with\n"
		"a server listening on MySQL standard port %d will be checked\n\n"
		"Options:\n"
		" -H, --master_hostname=STRING or IPADDRESS\n"
		"   Check master server on the indicated host\n"
		" -S, --slave_hostname=STRING or IPADDRESS\n"
		"   Check slave server on the indicated host\n"
		" -l, --limit=INTEGER\n"
		"   Maximum threshold which the slave can be behind the master in bytes\n"
		" -P, --port=INTEGER\n"
		"   Make connection on the indicated port\n"
		" -u, --username=STRING\n"
		"   Connect using the indicated username\n"
		"     Configuration note: the user under which the service runs needs "
        "     GRANT process ON *.* TO username@\"%\"\n"
		" -p, --password=STRING\n"
		"   Use the indicated password to authenticate the connection\n"
		"   ==> IMPORTANT: THIS FORM OF AUTHENTICATION IS NOT SECURE!!! <==\n"
		"   Your clear-text password will be visible as a process table entry\n"
		" -h, --help\n"
		"    Print detailed help screen\n"
		" -v, --verbose\n"
		"    Print all information\n"
		" -V, --version\n" "    Print version information\n\n", MYSQL_PORT);

	support ();
}





void
print_usage (void)
{
	printf
		("Usage: %s [-v] [-H masterhost] [-S slavehost] [-P port] [-u user] "
		"[-p password] [-l limit]\n"
		"       %s --help\n"
		"       %s --version\n", PROGNAME, PROGNAME, PROGNAME);
}
